#ifndef MALACHSCRIPT_LEXTOKEN_HPP
#define MALACHSCRIPT_LEXTOKEN_HPP

#include <memory>
#include "../../CoreData/Identifier.hpp"
#include "../../TextSpan.hpp"
#include "LexTokenKind.hpp"

namespace MalachScript::Parser {
    class LexToken {
        std::unique_ptr<const LexToken> _next;
        TextSpan _span;

    public:
        inline LexToken(const TextSpan& span) : _span(span) {}
        virtual ~LexToken() = default;
        [[nodiscard]] virtual LexTokenKind GetKind() const noexcept = 0;
        [[nodiscard]] inline const std::unique_ptr<const LexToken>& GetNext() const noexcept { return _next; }
        [[nodiscard]] inline const TextSpan& GetSpan() const noexcept { return _span; }

        inline void SetNext(LexToken* token) { _next = std::unique_ptr<const LexToken>(token); }
    };

    template <LexTokenKind kind> class LexTokenImpl : public LexToken {
    public:
        inline LexTokenImpl(const TextSpan& span) : LexToken(span){};
        [[nodiscard]] inline LexTokenKind GetKind() const noexcept final { return kind; }
    };

    class IntegerLiteral : public LexTokenImpl<LexTokenKind::IntegerLiteral> {
        uint64_t _value;

    public:
        IntegerLiteral(const TextSpan& span, uint64_t value)
            : LexTokenImpl<LexTokenKind::IntegerLiteral>(span), _value(value) {}
        [[nodiscard]] inline uint64_t GetValue() const noexcept { return _value; }
    };

    class FloatLiteral : public LexTokenImpl<LexTokenKind::FloatLiteral> {
        double _value;

    public:
        FloatLiteral(const TextSpan& span, double value)
            : LexTokenImpl<LexTokenKind::FloatLiteral>(span), _value(value) {}
        [[nodiscard]] inline double GetValue() const noexcept { return _value; }
    };

    class StringLiteral : public LexTokenImpl<LexTokenKind::StringLiteral> {
        std::u8string _value;

    public:
        StringLiteral(const TextSpan& span, std::u8string value)
            : LexTokenImpl<LexTokenKind::StringLiteral>(span), _value(std::move(value)) {}
        [[nodiscard]] inline const std::u8string& GetValue() const noexcept { return _value; }
    };

    class IdentifierToken : public LexTokenImpl<LexTokenKind::Identifier> {
        Identifier _value;

    public:
        IdentifierToken(const TextSpan& span, Identifier value)
            : LexTokenImpl<LexTokenKind::Identifier>(span), _value(value) {}
        [[nodiscard]] inline const Identifier& GetValue() const noexcept { return _value; }
    };
}

#endif // MALACHSCRIPT_LEXTOKEN_HPP
