#ifndef MALACHSCRIPT_PARSEDSTATEMENT_HPP
#define MALACHSCRIPT_PARSEDSTATEMENT_HPP

#include <utility>
#include <vector>
#include "../../CoreData/AccessModifier.hpp"
#include "../../CoreData/FuncAttr.hpp"
#include "../../CoreData/TypeMod.hpp"
#include "../../TextSpan.hpp"
#include "../Expressions/ParsedExpression.hpp"
#include "ParsedStatementKind.hpp"
namespace MalachScript::Parser {
    class ParsedStatement {
        TextSpan _span;

    public:
        ParsedStatement(TextSpan span) : _span(span) {}
        virtual ~ParsedStatement() = default;
        [[nodiscard]] virtual ParsedStatementKind GetKind() const noexcept = 0;
        [[nodiscard]] inline const TextSpan& GetSpan() const noexcept { return _span; }
    };

    template <ParsedStatementKind kind> class ParsedStatementImpl : public ParsedStatement {
    public:
        ParsedStatementImpl(TextSpan span) : ParsedStatement(span) {}
        [[nodiscard]] inline ParsedStatementKind GetKind() const noexcept final { return kind; }
    };

    class ParsedScriptStatement : public ParsedStatementImpl<ParsedStatementKind::Script> {
        std::vector<std::unique_ptr<const ParsedStatement>> _statements;

    public:
        ParsedScriptStatement(TextSpan span, const std::vector<const ParsedStatement*>& statements)
            : ParsedStatementImpl<ParsedStatementKind::Script>(span), _statements(statements.size()) {
            for (size_t i = 0; i < statements.size(); i++)
                _statements[i] = std::unique_ptr<const ParsedStatement>(statements[i]);
        }

        [[nodiscard]] inline const std::vector<std::unique_ptr<const ParsedStatement>>& GetStatements() const noexcept {
            return _statements;
        }
    };

    class ParsedClassStatement : public ParsedStatementImpl<ParsedStatementKind::Class> {
        Identifier _identifier;
        std::vector<Identifier> _inherits;
        std::vector<std::unique_ptr<const ParsedStatement>> _body;

    public:
        ParsedClassStatement(TextSpan span, Identifier identifier, std::vector<Identifier> inherits,
                             const std::vector<const ParsedStatement*>& body)
            : ParsedStatementImpl<ParsedStatementKind::Class>(span), _identifier(identifier), _inherits(inherits),
              _body(body.size()) {
            for (size_t i = 0; i < body.size(); i++)
                _body[i] = std::unique_ptr<const ParsedStatement>(body[i]);
        }

        [[nodiscard]] const Identifier& GetIdentifier() const noexcept { return _identifier; }
        [[nodiscard]] const std::vector<Identifier>& GetInherits() const noexcept { return _inherits; }
        [[nodiscard]] const std::vector<std::unique_ptr<const ParsedStatement>>& GetBody() const noexcept {
            return _body;
        }
    };

    class ParsedTypeDefStatement : public ParsedStatementImpl<ParsedStatementKind::TypeDef> {
        Identifier _defineFrom;
        Identifier _defineTo;

    public:
        ParsedTypeDefStatement(TextSpan span, const Identifier& defineFrom, const Identifier& defineTo)
            : ParsedStatementImpl<ParsedStatementKind::TypeDef>(span), _defineFrom(defineFrom), _defineTo(defineTo) {}

        [[nodiscard]] inline const Identifier& GetDefineFrom() const noexcept { return _defineFrom; }
        [[nodiscard]] inline const Identifier& GetDefineTo() const noexcept { return _defineTo; }
    };

    class ParsedNamespaceStatement : public ParsedStatementImpl<ParsedStatementKind::Namespace> {
        Identifier _identifier;
        std::unique_ptr<const ParsedStatement> _parsedScript;

    public:
        ParsedNamespaceStatement(TextSpan span, const Identifier& identifier, const ParsedScriptStatement* script)
            : ParsedStatementImpl<ParsedStatementKind::Namespace>(span), _identifier(identifier),
              _parsedScript(script) {}

        [[nodiscard]] inline const Identifier& GetIdentifier() const noexcept { return _identifier; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetScript() const noexcept {
            return _parsedScript;
        }
    };

    class ParsedTypeStatement : public ParsedStatementImpl<ParsedStatementKind::Type> {
        bool _isConst;
        bool _isArray;
        bool _isHandle;
        ScopedIdentifier _scopedIdentifier;

    public:
        ParsedTypeStatement(TextSpan span, bool isConst, bool isArray, bool isHandle,
                            const ScopedIdentifier& scopedIdentifier)
            : ParsedStatementImpl<ParsedStatementKind::Type>(span), _isConst(isConst), _isArray(isArray),
              _isHandle(isHandle), _scopedIdentifier(scopedIdentifier) {}

        [[nodiscard]] inline bool IsConst() const noexcept { return _isConst; }
        [[nodiscard]] inline bool IsArray() const noexcept { return _isArray; }
        [[nodiscard]] inline bool IsHandle() const noexcept { return _isHandle; }
        [[nodiscard]] inline const ScopedIdentifier& GetScopedIdentifier() const noexcept { return _scopedIdentifier; }
    };

    class ParsedParamListStatement : public ParsedStatementImpl<ParsedStatementKind::ParamList> {
    public:
        class ParsedParameter {
        private:
            std::unique_ptr<const ParsedTypeStatement> _typeStatement = nullptr;
            TypeMod _typeMod = TypeMod::None;
            Identifier _identifier;
            std::unique_ptr<const ParsedExpression> _defaultExpression = nullptr;

        public:
            ParsedParameter(const ParsedTypeStatement* typeStatement, TypeMod typeMod, const Identifier& identifier,
                            const ParsedExpression* defaultExpression)
                : _typeStatement(typeStatement), _typeMod(typeMod), _identifier(identifier),
                  _defaultExpression(defaultExpression){};

            [[nodiscard]] std::unique_ptr<const ParsedTypeStatement>& GetTypeStatement() noexcept {
                return _typeStatement;
            }
            [[nodiscard]] const std::unique_ptr<const ParsedTypeStatement>& GetTypeStatement() const noexcept {
                return _typeStatement;
            }

            [[nodiscard]] TypeMod& GetTypeMod() noexcept { return _typeMod; }
            [[nodiscard]] const TypeMod& GetTypeMod() const noexcept { return _typeMod; }

            [[nodiscard]] Identifier& GetIdentifier() noexcept { return _identifier; }
            [[nodiscard]] const Identifier& GetIdentifier() const noexcept { return _identifier; }

            [[nodiscard]] std::unique_ptr<const ParsedExpression>& GetDefaultExpression() noexcept {
                return _defaultExpression;
            }
            [[nodiscard]] const std::unique_ptr<const ParsedExpression>& GetDefaultExpression() const noexcept {
                return _defaultExpression;
            }
        };

    private:
        std::vector<std::unique_ptr<const ParsedParameter>> _parameters;

    public:
        ParsedParamListStatement(TextSpan span, const std::vector<const ParsedParameter*>& parameters)
            : ParsedStatementImpl<ParsedStatementKind::ParamList>(span), _parameters(parameters.size()) {
            for (size_t i = 0; i < parameters.size(); i++) {
                _parameters[i] = std::unique_ptr<const ParsedParameter>(parameters[i]);
            }
        };

        [[nodiscard]] const std::vector<std::unique_ptr<const ParsedParameter>>& GetParameters() const noexcept {
            return _parameters;
        }
    };

    class ParsedFuncStatement : public ParsedStatementImpl<ParsedStatementKind::Func> {
    private:
        bool _isShared;
        bool _isExternal;
        AccessModifier _access;
        std::unique_ptr<const ParsedStatement> _type;
        bool _returnsReference;
        Identifier _identifier;
        std::unique_ptr<const ParsedStatement> _paramList;
        bool _isConst;
        FuncAttr _funcAttr;
        std::unique_ptr<const ParsedStatement> _statBlock;

    public:
        ParsedFuncStatement(const TextSpan& span, bool isShared, bool isExternal, AccessModifier access,
                            const ParsedStatement* type, bool returnsReference, const Identifier& identifier,
                            const ParsedStatement* paramList, bool isConst, FuncAttr funcAttr,
                            const ParsedStatement* statBlock)
            : ParsedStatementImpl<ParsedStatementKind::Func>(span), _isShared(isShared), _isExternal(isExternal),
              _access(access), _type(type), _returnsReference(returnsReference), _identifier(identifier),
              _paramList(paramList), _isConst(isConst), _funcAttr(funcAttr), _statBlock(statBlock) {}

        [[nodiscard]] inline bool IsShared() const noexcept { return _isShared; }
        [[nodiscard]] inline bool IsExternal() const noexcept { return _isExternal; }
        [[nodiscard]] inline AccessModifier GetAccess() const noexcept { return _access; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetTypeStatement() const noexcept {
            return _type;
        }
        [[nodiscard]] inline bool ReturnsReference() const noexcept { return _returnsReference; }
        [[nodiscard]] inline const Identifier& GetIdentifier() const noexcept { return _identifier; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetParamList() const noexcept {
            return _paramList;
        }
        [[nodiscard]] inline bool IsConst() const noexcept { return _isConst; }
        [[nodiscard]] inline FuncAttr GetFuncAttr() const noexcept { return _funcAttr; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetStatBlock() const noexcept {
            return _statBlock;
        }
    };

    class ParsedVirtPropStatement : public ParsedStatementImpl<ParsedStatementKind::VirtProp> {
    public:
        ParsedVirtPropStatement(const TextSpan& span, AccessModifier access, const ParsedStatement* returnType,
                                bool isReturnTypeRef, Identifier identifier, bool hasGet, bool getConst,
                                FuncAttr getAttr, const ParsedStatement* getStatement, bool hasSet, bool setConst,
                                FuncAttr setAttr, const ParsedStatement* setStatement)
            : ParsedStatementImpl(span), _access(access), _returnType(returnType), _isReturnTypeRef(isReturnTypeRef),
              _identifier(identifier), _hasGet(hasGet), _getConst(getConst), _getAttr(getAttr),
              _getStatement(getStatement), _hasSet(hasSet), _setConst(setConst), _setAttr(setAttr),
              _setStatement(setStatement) {}

        [[nodiscard]] inline AccessModifier GetAccess() const noexcept { return _access; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetReturnType() const noexcept {
            return _returnType;
        }
        [[nodiscard]] inline bool IsReturnTypeRef() const noexcept { return _isReturnTypeRef; }
        [[nodiscard]] inline const Identifier& GetIdentifier() const noexcept { return _identifier; }

        [[nodiscard]] inline bool HasGet() const noexcept { return _hasGet; }
        [[nodiscard]] inline bool IsGetConst() const noexcept { return _getConst; }
        [[nodiscard]] inline FuncAttr GetGetFuncAttr() const noexcept { return _getAttr; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetGetStatement() const noexcept {
            return _getStatement;
        }

        [[nodiscard]] inline bool HasSet() const noexcept { return _hasSet; }
        [[nodiscard]] inline bool IsSetConst() const noexcept { return _setConst; }
        [[nodiscard]] inline FuncAttr GetSetFuncAttr() const noexcept { return _setAttr; }
        [[nodiscard]] inline const std::unique_ptr<const ParsedStatement>& GetSetStatement() const noexcept {
            return _setStatement;
        }

    private:
        AccessModifier _access;
        std::unique_ptr<const ParsedStatement> _returnType;
        bool _isReturnTypeRef;
        Identifier _identifier;

        bool _hasGet = false;
        bool _getConst = false;
        FuncAttr _getAttr = FuncAttr::None;
        std::unique_ptr<const ParsedStatement> _getStatement;
        bool _hasSet = false;
        bool _setConst = false;
        FuncAttr _setAttr = FuncAttr::None;
        std::unique_ptr<const ParsedStatement> _setStatement;
    };
}

#endif // MALACHSCRIPT_PARSEDSTATEMENT_HPP
